/*!
 * Tozny e3db
 *
 * LICENSE
 *
 * Tozny dual licenses this product. For commercial use, please contact
 * info@tozny.com. For non-commercial use, the contents of this file are
 * subject to the TOZNY NON-COMMERCIAL LICENSE (the "License") which
 * permits use of the software only by government agencies, schools,
 * universities, non-profit organizations or individuals on projects that
 * do not receive external funding other than government research grants
 * and contracts.  Any other use requires a commercial license. You may
 * not use this file except in compliance with the License. You may obtain
 * a copy of the License at https://tozny.com/legal/non-commercial-license.
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License. Portions of the software are Copyright (c) TOZNY LLC, 2018-19.
 * All rights reserved.
 *
 * @copyright Copyright (c) 2018-19 Tozny, LLC (https://tozny.com)
 */

/* global fetch */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

var _getIterator2 = require('babel-runtime/core-js/get-iterator');

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _stringify = require('babel-runtime/core-js/json/stringify');

var _stringify2 = _interopRequireDefault(_stringify);

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

/* eslint-enable */

/**
 * Potentially refresh the authorization token used during requests to the E3DB server.
 *
 * The token will be cached for 10 minutes before being automatically refreshed.
 *
 * @param {Config} config E3DB client configuration
 *
 * @returns {Promise<string>}
 */
var getToken = function () {
  var _ref = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee(client) {
    var response, json;
    return _regenerator2.default.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            if (!(client._authToken === null || Date.now() > client._authTokenTimeout)) {
              _context.next = 17;
              break;
            }

            response = void 0;
            _context.prev = 2;
            _context.next = 5;
            return fetch(client.config.apiUrl + '/v1/auth/token', {
              method: 'POST',
              headers: {
                'Content-Type': 'x-www-form-urlencoded',
                Authorization: 'Basic ' + btoa(client.config.apiKeyId + ':' + client.config.apiSecret)
              },
              body: 'grant_type=client_credentials'
            });

          case 5:
            response = _context.sent;
            _context.next = 11;
            break;

          case 8:
            _context.prev = 8;
            _context.t0 = _context['catch'](2);

            console.log('ERROR>>>', _context.t0);

          case 11:
            _context.next = 13;
            return response.json();

          case 13:
            json = _context.sent;

            client._authToken = json.access_token;
            client._authTokenTimeout = Date.parse(json.expires_at);

            return _context.abrupt('return', client._authToken);

          case 17:
            return _context.abrupt('return', _promise2.default.resolve(client._authToken));

          case 18:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[2, 8]]);
  }));

  return function getToken(_x) {
    return _ref.apply(this, arguments);
  };
}();

/**
 * Transparent fetch() wrapper to set up OAuth2 authentication headers
 *
 * @param {Client} client  E3DB client instance
 * @param {string} url     Absolute URL to fetch from the server
 * @param {object} options Object representing additional settings for the fetch
 *
 * @returns {Promise}
 */


var oauthFetch = function () {
  var _ref2 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee2(client, url, options) {
    var token;
    return _regenerator2.default.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.next = 2;
            return getToken(client);

          case 2:
            token = _context2.sent;

            options.headers = options.headers || {};
            options.headers.Authorization = 'Bearer ' + token;
            return _context2.abrupt('return', fetch(url, options));

          case 6:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function oauthFetch(_x2, _x3, _x4) {
    return _ref2.apply(this, arguments);
  };
}();

/**
 * Check the return status of a fetch request and throw an error if one occurred
 *
 * @param {Response} response
 *
 * @returns {Promise}
 */


var checkStatus = function () {
  var _ref3 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee3(response) {
    var error;
    return _regenerator2.default.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            if (!(response.status >= 200 && response.status < 300)) {
              _context3.next = 2;
              break;
            }

            return _context3.abrupt('return', _promise2.default.resolve(response));

          case 2:
            error = new Error(response.statusText);

            error.response = response;
            throw error;

          case 5:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function checkStatus(_x5) {
    return _ref3.apply(this, arguments);
  };
}();

/**
 * Retrieve an access key from the server.
 *
 * @param {Client} client E3DB client instance
 * @param {string} writerId Writer/Authorizer for the access key
 * @param {string} userId   Record subject
 * @param {string} readerId Authorized reader
 * @param {string} type     Record type for which the key will be used
 *
 * @returns {Promise<EAKInfo|null>} Encrypted access key on success, NULL if no key exists.
 */


var getEncryptedAccessKey = function () {
  var _ref4 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee4(client, writerId, userId, readerId, type) {
    var response;
    return _regenerator2.default.wrap(function _callee4$(_context4) {
      while (1) {
        switch (_context4.prev = _context4.next) {
          case 0:
            _context4.next = 2;
            return oauthFetch(client, client.config.apiUrl + '/v1/storage/access_keys/' + writerId + '/' + userId + '/' + readerId + '/' + type, {
              method: 'GET',
              headers: {
                'Content-Type': 'application/json'
              }
            });

          case 2:
            response = _context4.sent;

            if (!(response.status && response.status === 404)) {
              _context4.next = 5;
              break;
            }

            return _context4.abrupt('return', _promise2.default.resolve(null));

          case 5:
            return _context4.abrupt('return', checkStatus(response).then(function (response) {
              return response.json();
            }).then(function (eak) {
              return _eakInfo2.default.decode(eak);
            }));

          case 6:
          case 'end':
            return _context4.stop();
        }
      }
    }, _callee4, this);
  }));

  return function getEncryptedAccessKey(_x6, _x7, _x8, _x9, _x10) {
    return _ref4.apply(this, arguments);
  };
}();

/**
 * Retrieve an access key from the server.
 *
 * @param {Client} client E3DB client instance
 * @param {string} writerId Writer/Authorizer for the access key
 * @param {string} userId   Record subject
 * @param {string} readerId Authorized reader
 * @param {string} type     Record type for which the key will be used
 *
 * @returns {Promise<string|null>} Decrypted access key on success, NULL if no key exists.
 */


var getAccessKey = function () {
  var _ref5 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee5(client, writerId, userId, readerId, type, crypto) {
    var cacheKey;
    return _regenerator2.default.wrap(function _callee5$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            cacheKey = writerId + '.' + userId + '.' + type;

            if (!(client._akCache[cacheKey] !== undefined)) {
              _context5.next = 3;
              break;
            }

            return _context5.abrupt('return', _promise2.default.resolve(client._akCache[cacheKey]));

          case 3:
            return _context5.abrupt('return', getEncryptedAccessKey(client, writerId, userId, readerId, type).then(function (eak) {
              if (eak === null) {
                return _promise2.default.resolve(null);
              }
              return crypto.decryptEak(client.config.privateKey, eak);
            }).then(function (key) {
              if (key !== null) {
                client._akCache[cacheKey] = key;
              }
              return _promise2.default.resolve(key);
            }).catch(function (err) {
              return console.log(err);
            }));

          case 4:
          case 'end':
            return _context5.stop();
        }
      }
    }, _callee5, this);
  }));

  return function getAccessKey(_x11, _x12, _x13, _x14, _x15, _x16) {
    return _ref5.apply(this, arguments);
  };
}();

/**
 * Create an access key on the server.
 *
 * @param {Client} client   E3DB client instance
 * @param {string} writerId Writer/Authorizer for the access key
 * @param {string} userId   Record subject
 * @param {string} readerId Authorized reader
 * @param {string} type     Record type for which the key will be used
 * @param {string} ak       Unencrypted access key

 @returns {Promise<string>} Decrypted access key
 */


var putAccessKey = function () {
  var _ref6 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee6(client, writerId, userId, readerId, type, ak, crypto) {
    var clientInfo, readerKey, eak;
    return _regenerator2.default.wrap(function _callee6$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            _context6.next = 2;
            return client.getClient(readerId);

          case 2:
            clientInfo = _context6.sent;
            readerKey = clientInfo.publicKey.curve25519;
            _context6.next = 6;
            return crypto.encryptAk(client.config.privateKey, ak, readerKey);

          case 6:
            eak = _context6.sent;
            return _context6.abrupt('return', oauthFetch(client, client.config.apiUrl + '/v1/storage/access_keys/' + writerId + '/' + userId + '/' + readerId + '/' + type, {
              method: 'PUT',
              headers: {
                'Content-Type': 'application/json'
              },
              body: (0, _stringify2.default)({ eak: eak })
            }).then(checkStatus).then(function () {
              var cacheKey = writerId + '.' + userId + '.' + type;
              client._akCache[cacheKey] = ak;

              return _promise2.default.resolve(ak);
            }));

          case 8:
          case 'end':
            return _context6.stop();
        }
      }
    }, _callee6, this);
  }));

  return function putAccessKey(_x17, _x18, _x19, _x20, _x21, _x22, _x23) {
    return _ref6.apply(this, arguments);
  };
}();

/**
 * Delete an access key on the server.
 *
 * @param {Client} client   E3DB client instance
 * @param {string} writerId Writer/Authorizer for the access key
 * @param {string} userId   Record subject
 * @param {string} readerId Authorized reader
 * @param {string} type     Record type for which the key will be used
 *
 * @returns {Promise<bool>}
 */


var deleteAccessKey = function () {
  var _ref7 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee7(client, writerId, userId, readerId, type) {
    var request, cacheKey;
    return _regenerator2.default.wrap(function _callee7$(_context7) {
      while (1) {
        switch (_context7.prev = _context7.next) {
          case 0:
            _context7.next = 2;
            return oauthFetch(client, client.config.apiUrl + '/v1/storage/access_keys/' + writerId + '/' + userId + '/' + readerId + '/' + type, {
              method: 'DELETE',
              headers: {
                'Content-Type': 'application/json'
              }
            });

          case 2:
            request = _context7.sent;
            _context7.next = 5;
            return checkStatus(request);

          case 5:
            cacheKey = writerId + '.' + userId + '.' + type;

            delete client._akCache[cacheKey];

            return _context7.abrupt('return', true);

          case 8:
          case 'end':
            return _context7.stop();
        }
      }
    }, _callee7, this);
  }));

  return function deleteAccessKey(_x24, _x25, _x26, _x27, _x28) {
    return _ref7.apply(this, arguments);
  };
}();

/**
 * Fetch the access key for a record type and use it to decrypt a given record.
 *
 * @param {Client} client E3DB client instance
 * @param {Record} encrypted Record to be decrypted
 *
 * @return {Promise<Record>}
 */


var decryptRecord = function () {
  var _ref8 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee8(client, encrypted, crypto) {
    var ak;
    return _regenerator2.default.wrap(function _callee8$(_context8) {
      while (1) {
        switch (_context8.prev = _context8.next) {
          case 0:
            _context8.next = 2;
            return getAccessKey(client, encrypted.meta.writerId, encrypted.meta.userId, client.config.clientId, encrypted.meta.type, crypto);

          case 2:
            ak = _context8.sent;

            if (!(ak === null)) {
              _context8.next = 5;
              break;
            }

            throw new Error('No access key available.');

          case 5:
            return _context8.abrupt('return', crypto.decryptRecord(encrypted, ak));

          case 6:
          case 'end':
            return _context8.stop();
        }
      }
    }, _callee8, this);
  }));

  return function decryptRecord(_x29, _x30, _x31) {
    return _ref8.apply(this, arguments);
  };
}();

/**
 * Fetch the access key for a record type and use it to encrypt a given record.
 *
 * @param {Client} client E3DB client instance
 * @param {Record} record Record to be decrypted
 *
 * @return {Promise<Record>}
 */


var encryptRecord = function () {
  var _ref9 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee9(client, record, crypto) {
    var ak;
    return _regenerator2.default.wrap(function _callee9$(_context9) {
      while (1) {
        switch (_context9.prev = _context9.next) {
          case 0:
            _context9.next = 2;
            return getAccessKey(client, record.meta.writerId, record.meta.userId, client.config.clientId, record.meta.type, crypto);

          case 2:
            ak = _context9.sent;

            if (!(ak === null)) {
              _context9.next = 9;
              break;
            }

            _context9.next = 6;
            return crypto.randomKey();

          case 6:
            ak = _context9.sent;
            _context9.next = 9;
            return putAccessKey(client, record.meta.writerId, record.meta.userId, client.config.clientId, record.meta.type, ak, crypto);

          case 9:
            return _context9.abrupt('return', crypto.encryptRecord(record, ak));

          case 10:
          case 'end':
            return _context9.stop();
        }
      }
    }, _callee9, this);
  }));

  return function encryptRecord(_x32, _x33, _x34) {
    return _ref9.apply(this, arguments);
  };
}();

/**
 * Core client module used to interact with the E3DB API.
 *
 * @property {Config} config E3DB client configuration.
 */

// *******************************************************************


require('es6-promise/auto');

require('isomorphic-fetch');

var _crypto = require('./crypto');

var _crypto2 = _interopRequireDefault(_crypto);

var _config = require('./config');

var _config2 = _interopRequireDefault(_config);

var _clientDetails = require('./types/clientDetails');

var _clientDetails2 = _interopRequireDefault(_clientDetails);

var _clientInfo = require('./types/clientInfo');

var _clientInfo2 = _interopRequireDefault(_clientInfo);

var _eakInfo = require('./types/eakInfo');

var _eakInfo2 = _interopRequireDefault(_eakInfo);

var _file = require('./types/file');

var _file2 = _interopRequireDefault(_file);

var _incomingSharingPolicy = require('./types/incomingSharingPolicy');

var _incomingSharingPolicy2 = _interopRequireDefault(_incomingSharingPolicy);

var _meta = require('./types/meta');

var _meta2 = _interopRequireDefault(_meta);

var _outgoingSharingPolicy = require('./types/outgoingSharingPolicy');

var _outgoingSharingPolicy2 = _interopRequireDefault(_outgoingSharingPolicy);

var _publicKey = require('./types/publicKey');

var _publicKey2 = _interopRequireDefault(_publicKey);

var _query2 = require('./types/query');

var _query3 = _interopRequireDefault(_query2);

var _queryResult = require('./types/queryResult');

var _queryResult2 = _interopRequireDefault(_queryResult);

var _record = require('./types/record');

var _record2 = _interopRequireDefault(_record);

var _recordData = require('./types/recordData');

var _recordData2 = _interopRequireDefault(_recordData);

var _recordInfo = require('./types/recordInfo');

var _recordInfo2 = _interopRequireDefault(_recordInfo);

var _signedDocument = require('./types/signedDocument');

var _signedDocument2 = _interopRequireDefault(_signedDocument);

var _signingKey = require('./types/signingKey');

var _signingKey2 = _interopRequireDefault(_signingKey);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var DEFAULT_QUERY_COUNT = 100;
var DEFAULT_API_URL = 'https://api.e3db.com';
var EMAIL = /(.+)@(.+){2,}\.(.+){2,}/;

/* eslint-disable no-buffer-constructor */

/**
 * Fallback polyfill to allow for HTTP Basic authentication from either Node
 * or browser-based JavaScript.
 *
 * @param {string} str String to encode as Base64
 */
var btoa = function btoa(str) {
  return new Buffer(str).toString('base64');
};var PreClient = function PreClient(crypto) {
  if (!(crypto instanceof _crypto2.default)) {
    throw new Error('Must pass a valid Crypto object');
  }

  var Client = function () {
    (0, _createClass3.default)(Client, null, [{
      key: 'register',

      /**
       * Register a new client with a specific account.
       *
       * @param {string}  registrationToken Registration token as presented by the admin console
       * @param {string}  clientName        Distinguishable name to be used for the token in the console
       * @param {KeyPair} cryptoKeys        Curve25519 keypair used for encryption
       * @param {KeyPair} signingKeys       Ed25519 keypair used for signing
       * @param {bool}    [backup]          Optional flag to automatically back up the newly-created credentials to the account service
       * @param {string}  [apiUrl]          Base URI for the e3DB API
       *
       * @returns {ClientDetails}
       */
      value: function () {
        var _ref10 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee10(registrationToken, clientName, cryptoKeys, signingKeys) {
          var backup = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : false;
          var apiUrl = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : DEFAULT_API_URL;
          var payload, backupClientId, request, response, json, details, config, client;
          return _regenerator2.default.wrap(function _callee10$(_context10) {
            while (1) {
              switch (_context10.prev = _context10.next) {
                case 0:
                  /* eslint-disable camelcase */
                  payload = void 0;

                  if (signingKeys) {
                    payload = {
                      token: registrationToken,
                      client: {
                        name: clientName,
                        public_key: new _publicKey2.default(cryptoKeys.publicKey),
                        signing_key: new _signingKey2.default(signingKeys.publicKey)
                      }
                    };
                  } else {
                    payload = {
                      token: registrationToken,
                      client: {
                        name: clientName,
                        public_key: new _publicKey2.default(cryptoKeys.publicKey)
                      }
                    };
                  }
                  /* eslint-enable */
                  backupClientId = false;
                  _context10.next = 5;
                  return fetch(apiUrl + '/v1/account/e3db/clients/register', {
                    method: 'POST',
                    headers: {
                      'Content-Type': 'application/json'
                    },
                    body: (0, _stringify2.default)(payload)
                  });

                case 5:
                  request = _context10.sent;
                  _context10.next = 8;
                  return checkStatus(request);

                case 8:
                  response = _context10.sent;

                  if (response.headers.has('X-Backup-Client')) {
                    backupClientId = response.headers.get('X-Backup-Client');
                  }
                  _context10.next = 12;
                  return response.json();

                case 12:
                  json = _context10.sent;
                  _context10.next = 15;
                  return _clientDetails2.default.decode(json);

                case 15:
                  details = _context10.sent;

                  if (!(backup && backupClientId)) {
                    _context10.next = 23;
                    break;
                  }

                  if (!(cryptoKeys.privateKey === null)) {
                    _context10.next = 19;
                    break;
                  }

                  throw new Error('Cannot back up credentials without a private key!');

                case 19:
                  config = signingKeys ? new _config2.default(details.clientId, details.apiKeyId, details.apiSecret, cryptoKeys.publicKey, cryptoKeys.privateKey, apiUrl, signingKeys.publicKey, signingKeys.privateKey) : new _config2.default(details.clientId, details.apiKeyId, details.apiSecret, cryptoKeys.publicKey, cryptoKeys.privateKey, apiUrl);
                  // This client may not have crypto... because it doesn't have the PreClient

                  client = new Client(config);
                  _context10.next = 23;
                  return client.backup(backupClientId, registrationToken);

                case 23:
                  return _context10.abrupt('return', _promise2.default.resolve(details));

                case 24:
                case 'end':
                  return _context10.stop();
              }
            }
          }, _callee10, this);
        }));

        function register(_x37, _x38, _x39, _x40) {
          return _ref10.apply(this, arguments);
        }

        return register;
      }()

      /**
       * Proxy to generating a new key pair using the crypto module provided.
       *
       * @returns {KeyPair} Base64URL-encoded representation of the new keypair
       */

    }, {
      key: 'generateKeypair',
      value: function () {
        var _ref11 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee11() {
          return _regenerator2.default.wrap(function _callee11$(_context11) {
            while (1) {
              switch (_context11.prev = _context11.next) {
                case 0:
                  return _context11.abrupt('return', crypto.generateKeypair());

                case 1:
                case 'end':
                  return _context11.stop();
              }
            }
          }, _callee11, this);
        }));

        function generateKeypair() {
          return _ref11.apply(this, arguments);
        }

        return generateKeypair;
      }()

      /**
       * Proxy to generating a new signing key pair using the crypto module provided.
       *
       * @returns {KeyPair} Base64URL-encoded representation of the new keypair
       */

    }, {
      key: 'generateSigningKeypair',
      value: function () {
        var _ref12 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee12() {
          return _regenerator2.default.wrap(function _callee12$(_context12) {
            while (1) {
              switch (_context12.prev = _context12.next) {
                case 0:
                  return _context12.abrupt('return', crypto.generateSigningKeypair());

                case 1:
                case 'end':
                  return _context12.stop();
              }
            }
          }, _callee12, this);
        }));

        function generateSigningKeypair() {
          return _ref12.apply(this, arguments);
        }

        return generateSigningKeypair;
      }()
    }]);

    function Client(config) {
      (0, _classCallCheck3.default)(this, Client);

      if (!(config instanceof _config2.default)) {
        throw new Error('Config must be a valid Config object');
      }
      this.config = config;
      this._authToken = null;
      this._authTokenTimeout = 0; // Minimum UNIX timestamp
      this._akCache = {};
    }

    /**
     * Get an access key from the cache if it exists, otherwise decrypt
     * the provided EAK and populate the cache.
     *
     * @param {string}  writerId
     * @param {string}  userId
     * @param {string}  readerId
     * @param {string}  type
     * @param {EAKInfo} eak
     *
     * @returns {Promise<string>}
     */


    (0, _createClass3.default)(Client, [{
      key: '_getCachedAk',
      value: function () {
        var _ref13 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee13(writerId, userId, readerId, type, eak) {
          var cacheKey, ak;
          return _regenerator2.default.wrap(function _callee13$(_context13) {
            while (1) {
              switch (_context13.prev = _context13.next) {
                case 0:
                  cacheKey = writerId + '.' + userId + '.' + type;
                  ak = this._akCache[cacheKey];

                  if (!(ak === undefined)) {
                    _context13.next = 7;
                    break;
                  }

                  _context13.next = 5;
                  return crypto.decryptEak(this.config.privateKey, eak);

                case 5:
                  ak = _context13.sent;

                  this._akCache[cacheKey] = ak;

                case 7:
                  return _context13.abrupt('return', _promise2.default.resolve(ak));

                case 8:
                case 'end':
                  return _context13.stop();
              }
            }
          }, _callee13, this);
        }));

        function _getCachedAk(_x41, _x42, _x43, _x44, _x45) {
          return _ref13.apply(this, arguments);
        }

        return _getCachedAk;
      }()

      /**
       * Get a client's information based on their ID.
       *
       * @param {string} clientId UUID of the client to fetch
       *
       * @returns {Promise<ClientInfo>}
       */

    }, {
      key: 'getClient',
      value: function () {
        var _ref14 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee14(clientId) {
          var request, response, json;
          return _regenerator2.default.wrap(function _callee14$(_context14) {
            while (1) {
              switch (_context14.prev = _context14.next) {
                case 0:
                  _context14.next = 2;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/clients/' + clientId, {
                    method: 'GET',
                    headers: {
                      'Content-Type': 'application/json'
                    }
                  });

                case 2:
                  request = _context14.sent;
                  _context14.next = 5;
                  return checkStatus(request);

                case 5:
                  response = _context14.sent;
                  _context14.next = 8;
                  return response.json();

                case 8:
                  json = _context14.sent;
                  return _context14.abrupt('return', _clientInfo2.default.decode(json));

                case 10:
                case 'end':
                  return _context14.stop();
              }
            }
          }, _callee14, this);
        }));

        function getClient(_x46) {
          return _ref14.apply(this, arguments);
        }

        return getClient;
      }()

      /**
       * Create a key for the current client as a writer if one does not exist
       * in the cache already. If no access key does exist, create a random one
       * and store it with the server.
       *
       * @param {string} type Record type for this key
       *
       * @returns {Promise<EAKInfo>}
       */

    }, {
      key: 'createWriterKey',
      value: function () {
        var _ref15 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee15(type) {
          var ak, eak;
          return _regenerator2.default.wrap(function _callee15$(_context15) {
            while (1) {
              switch (_context15.prev = _context15.next) {
                case 0:
                  _context15.next = 2;
                  return getAccessKey(this, this.config.clientId, this.config.clientId, this.config.clientId, type, crypto);

                case 2:
                  ak = _context15.sent;

                  if (!(ak === null)) {
                    _context15.next = 9;
                    break;
                  }

                  _context15.next = 6;
                  return crypto.randomKey();

                case 6:
                  ak = _context15.sent;
                  _context15.next = 9;
                  return putAccessKey(this, this.config.clientId, this.config.clientId, this.config.clientId, type, ak, crypto);

                case 9:
                  _context15.next = 11;
                  return crypto.encryptAk(this.config.privateKey, ak, this.config.publicKey);

                case 11:
                  eak = _context15.sent;
                  return _context15.abrupt('return', new _eakInfo2.default(eak, this.config.clientId, this.config.publicKey, this.config.clientId, this.config.publicSigningKey));

                case 13:
                case 'end':
                  return _context15.stop();
              }
            }
          }, _callee15, this);
        }));

        function createWriterKey(_x47) {
          return _ref15.apply(this, arguments);
        }

        return createWriterKey;
      }()

      /**
       * Get a key for the current client as the reader of a specific record written by someone else.
       *
       * @param {string} writerId Writer of the record in the database
       * @param {string} userID   Subject of the record in the database
       * @param {string} type     Type of record
       *
       * @returns {Promise<EAKInfo>}
       */

    }, {
      key: 'getReaderKey',
      value: function () {
        var _ref16 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee16(writerId, userId, type) {
          return _regenerator2.default.wrap(function _callee16$(_context16) {
            while (1) {
              switch (_context16.prev = _context16.next) {
                case 0:
                  return _context16.abrupt('return', getEncryptedAccessKey(this, writerId, userId, this.config.clientId, type, crypto));

                case 1:
                case 'end':
                  return _context16.stop();
              }
            }
          }, _callee16, this);
        }));

        function getReaderKey(_x48, _x49, _x50) {
          return _ref16.apply(this, arguments);
        }

        return getReaderKey;
      }()

      /**
       * Retrieve information about a client, primarily its UUID and public key,
       * based either on an already-known client ID or a discoverable client
       * email address.
       *
       * @param {string} clientId
       *
       * @returns {Promise<ClientInfo>}
       */

    }, {
      key: 'clientInfo',
      value: function () {
        var _ref17 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee17(clientId) {
          return _regenerator2.default.wrap(function _callee17$(_context17) {
            while (1) {
              switch (_context17.prev = _context17.next) {
                case 0:
                  if (!EMAIL.test(clientId)) {
                    _context17.next = 2;
                    break;
                  }

                  throw new Error('Client discovery by email address is not supported');

                case 2:
                  return _context17.abrupt('return', this.getClient(clientId));

                case 3:
                case 'end':
                  return _context17.stop();
              }
            }
          }, _callee17, this);
        }));

        function clientInfo(_x51) {
          return _ref17.apply(this, arguments);
        }

        return clientInfo;
      }()

      /**
       * Retrieve the Curve 25519 public key associated with a known client.
       *
       * @param {string} clientId
       *
       * @returns {Promise<PublicKey>}
       */

    }, {
      key: 'clientKey',
      value: function () {
        var _ref18 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee18(clientId) {
          var info;
          return _regenerator2.default.wrap(function _callee18$(_context18) {
            while (1) {
              switch (_context18.prev = _context18.next) {
                case 0:
                  if (!(clientId === this.clientId)) {
                    _context18.next = 2;
                    break;
                  }

                  return _context18.abrupt('return', _promise2.default.resolve(null));

                case 2:
                  _context18.next = 4;
                  return this.clientInfo(clientId);

                case 4:
                  info = _context18.sent;
                  return _context18.abrupt('return', _promise2.default.resolve(info.publicKey));

                case 6:
                case 'end':
                  return _context18.stop();
              }
            }
          }, _callee18, this);
        }));

        function clientKey(_x52) {
          return _ref18.apply(this, arguments);
        }

        return clientKey;
      }()

      /**
       * Reads a record from the E3DB system and decrypts it automatically.
       *
       * @param {string} recordId
       * @param {array}  [fields] Optional fields to select on the record
       *
       * @returns {Promise<Record>}
       */

    }, {
      key: 'read',
      value: function () {
        var _ref19 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee19(recordId) {
          var fields = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

          var path, mapped, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, field, request, response, json, record;

          return _regenerator2.default.wrap(function _callee19$(_context19) {
            while (1) {
              switch (_context19.prev = _context19.next) {
                case 0:
                  path = this.config.apiUrl + '/v1/storage/records/' + recordId;

                  if (!(fields !== null)) {
                    _context19.next = 23;
                    break;
                  }

                  mapped = [];
                  _iteratorNormalCompletion = true;
                  _didIteratorError = false;
                  _iteratorError = undefined;
                  _context19.prev = 6;

                  for (_iterator = (0, _getIterator3.default)(fields); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    field = _step.value;

                    mapped.push('field=' + field);
                  }

                  _context19.next = 14;
                  break;

                case 10:
                  _context19.prev = 10;
                  _context19.t0 = _context19['catch'](6);
                  _didIteratorError = true;
                  _iteratorError = _context19.t0;

                case 14:
                  _context19.prev = 14;
                  _context19.prev = 15;

                  if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                  }

                case 17:
                  _context19.prev = 17;

                  if (!_didIteratorError) {
                    _context19.next = 20;
                    break;
                  }

                  throw _iteratorError;

                case 20:
                  return _context19.finish(17);

                case 21:
                  return _context19.finish(14);

                case 22:
                  path += '?' + mapped.join('&');

                case 23:
                  _context19.next = 25;
                  return oauthFetch(this, path, {
                    method: 'GET',
                    headers: {
                      'Content-Type': 'application/json'
                    }
                  });

                case 25:
                  request = _context19.sent;
                  _context19.next = 28;
                  return checkStatus(request);

                case 28:
                  response = _context19.sent;
                  _context19.next = 31;
                  return response.json();

                case 31:
                  json = _context19.sent;
                  _context19.next = 34;
                  return _record2.default.decode(json);

                case 34:
                  record = _context19.sent;
                  return _context19.abrupt('return', decryptRecord(this, record, crypto));

                case 36:
                case 'end':
                  return _context19.stop();
              }
            }
          }, _callee19, this, [[6, 10, 14, 22], [15,, 17, 21]]);
        }));

        function read(_x54) {
          return _ref19.apply(this, arguments);
        }

        return read;
      }()

      /**
       * Create a new record entry with E3DB.
       *
       * @param {string} type  The content type with which to associate the record.
       * @param {object} data  A hashmap of the data to encrypt and store
       * @param {object} plain Optional hashmap of data to store with the record's meta in plaintext
       *
       * @return {Promise<Record>}
       */

    }, {
      key: 'write',
      value: function () {
        var _ref20 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee20(type, data) {
          var plain = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
          var meta, info, signature, record, encrypted;
          return _regenerator2.default.wrap(function _callee20$(_context20) {
            while (1) {
              switch (_context20.prev = _context20.next) {
                case 0:
                  // Build the record
                  if (data instanceof Object) {
                    data = new _recordData2.default(data);
                  }
                  meta = new _meta2.default(this.config.clientId, this.config.clientId, type, plain);
                  info = new _recordInfo2.default(meta, data);

                  if (!(this.config.version > 1)) {
                    _context20.next = 9;
                    break;
                  }

                  _context20.next = 6;
                  return this.sign(info);

                case 6:
                  _context20.t0 = _context20.sent;
                  _context20.next = 10;
                  break;

                case 9:
                  _context20.t0 = null;

                case 10:
                  signature = _context20.t0;
                  record = new _record2.default(meta, data, signature);
                  _context20.next = 14;
                  return encryptRecord(this, record, crypto);

                case 14:
                  encrypted = _context20.sent;
                  return _context20.abrupt('return', this.writeRaw(encrypted));

                case 16:
                case 'end':
                  return _context20.stop();
              }
            }
          }, _callee20, this);
        }));

        function write(_x56, _x57) {
          return _ref20.apply(this, arguments);
        }

        return write;
      }()

      /**
       * Write a previously stored encrypted/signed record directly to E3DB.
       *
       * @param {Record} record The fully-constructed record object, as returned by `encrypt()`
       *
       * @return {Promise<Record>}
       */

    }, {
      key: 'writeRaw',
      value: function () {
        var _ref21 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee21(record) {
          var request, response, json, written;
          return _regenerator2.default.wrap(function _callee21$(_context21) {
            while (1) {
              switch (_context21.prev = _context21.next) {
                case 0:
                  if (record instanceof _record2.default) {
                    _context21.next = 2;
                    break;
                  }

                  throw new Error('Can only write encrypted/signed records directly to the server!');

                case 2:
                  _context21.next = 4;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/records', {
                    method: 'POST',
                    headers: {
                      'Content-Type': 'application/json'
                    },
                    body: record.stringify()
                  });

                case 4:
                  request = _context21.sent;
                  _context21.next = 7;
                  return checkStatus(request);

                case 7:
                  response = _context21.sent;
                  _context21.next = 10;
                  return response.json();

                case 10:
                  json = _context21.sent;
                  _context21.next = 13;
                  return _record2.default.decode(json);

                case 13:
                  written = _context21.sent;
                  return _context21.abrupt('return', decryptRecord(this, written, crypto));

                case 15:
                case 'end':
                  return _context21.stop();
              }
            }
          }, _callee21, this);
        }));

        function writeRaw(_x58) {
          return _ref21.apply(this, arguments);
        }

        return writeRaw;
      }()

      /**
       * Encrypt a plaintext record using the AK wrapped and encrypted for the current
       * client. The key will be cached for future use.
       *
       * @param {string}            type  The content type with which to associate the record.
       * @param {RecordData|object} data  A hashmap of the data to encrypt and store
       * @param {object}            eak   Encrypted access key instance
       * @param {object}            plain Optional hashmap of data to store with the record's meta in plaintext
       *
       * @returns {Promise<Record>}
       */

    }, {
      key: 'localEncrypt',
      value: function () {
        var _ref22 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee22(type, data, eak) {
          var plain = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
          var ak, meta, recordInfo, signature, record;
          return _regenerator2.default.wrap(function _callee22$(_context22) {
            while (1) {
              switch (_context22.prev = _context22.next) {
                case 0:
                  _context22.next = 2;
                  return this._getCachedAk(this.config.clientId, this.config.clientId, this.config.clientId, type, eak);

                case 2:
                  ak = _context22.sent;


                  if (data instanceof Object) {
                    data = new _recordData2.default(data);
                  }

                  // Build the record
                  meta = new _meta2.default(this.config.clientId, this.config.clientId, type, plain);
                  recordInfo = new _recordInfo2.default(meta, data);

                  if (!(this.config.version > 1)) {
                    _context22.next = 12;
                    break;
                  }

                  _context22.next = 9;
                  return this.sign(recordInfo);

                case 9:
                  _context22.t0 = _context22.sent;
                  _context22.next = 13;
                  break;

                case 12:
                  _context22.t0 = null;

                case 13:
                  signature = _context22.t0;
                  record = new _record2.default(meta, data, signature);
                  return _context22.abrupt('return', crypto.encryptRecord(record, ak));

                case 16:
                case 'end':
                  return _context22.stop();
              }
            }
          }, _callee22, this);
        }));

        function localEncrypt(_x60, _x61, _x62) {
          return _ref22.apply(this, arguments);
        }

        return localEncrypt;
      }()

      /**
       * Sign a document and return the signature
       *
       * @param {Signable} document Serializable object to be signed.
       *
       * @returns {Promise<string>}
       */

    }, {
      key: 'sign',
      value: function () {
        var _ref23 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee23(document) {
          return _regenerator2.default.wrap(function _callee23$(_context23) {
            while (1) {
              switch (_context23.prev = _context23.next) {
                case 0:
                  if (!(this.config.version === 1)) {
                    _context23.next = 2;
                    break;
                  }

                  throw new Error('Cannot sign documents without a signing key!');

                case 2:
                  return _context23.abrupt('return', crypto.signDocument(document, this.config.privateSigningKey));

                case 3:
                case 'end':
                  return _context23.stop();
              }
            }
          }, _callee23, this);
        }));

        function sign(_x63) {
          return _ref23.apply(this, arguments);
        }

        return sign;
      }()

      /**
       * Decrypt an encrypted record using the AK wrapped and encrypted for the current
       * client. The key will be cached for future use.
       *
       * @param {Record}  record Record instance with encrypted data for decryption
       * @param {EAKInfo} eak    Encrypted access key instance
       *
       * @returns {Promise<Record>}
       */

    }, {
      key: 'localDecrypt',
      value: function () {
        var _ref24 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee24(record, eak) {
          var ak, decrypted, info, signed, verify;
          return _regenerator2.default.wrap(function _callee24$(_context24) {
            while (1) {
              switch (_context24.prev = _context24.next) {
                case 0:
                  if (!(eak.signerSigningKey === null)) {
                    _context24.next = 2;
                    break;
                  }

                  throw new Error('EAKInfo has no signing key!');

                case 2:
                  _context24.next = 4;
                  return this._getCachedAk(record.meta.writerId, record.meta.userId, this.config.clientId, record.meta.type, eak);

                case 4:
                  ak = _context24.sent;
                  _context24.next = 7;
                  return crypto.decryptRecord(record, ak);

                case 7:
                  decrypted = _context24.sent;
                  info = new _recordInfo2.default(decrypted.meta, decrypted.data);
                  signed = new _signedDocument2.default(info, decrypted.signature);
                  _context24.next = 12;
                  return this.verify(signed, eak.signerSigningKey.ed25519);

                case 12:
                  verify = _context24.sent;

                  if (verify) {
                    _context24.next = 15;
                    break;
                  }

                  throw new Error('Document failed verification');

                case 15:
                  return _context24.abrupt('return', decrypted);

                case 16:
                case 'end':
                  return _context24.stop();
              }
            }
          }, _callee24, this);
        }));

        function localDecrypt(_x64, _x65) {
          return _ref24.apply(this, arguments);
        }

        return localDecrypt;
      }()

      /**
       * Verify the signature attached to a specific document.
       *
       * @param {SignedDocument} signed        Document with an attached signature
       * @param {string}         publicSignKey Key to use during signature verification
       *
       * @returns {Promise<bool>}
       */

    }, {
      key: 'verify',
      value: function () {
        var _ref25 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee25(signed, publicSignKey) {
          var verified;
          return _regenerator2.default.wrap(function _callee25$(_context25) {
            while (1) {
              switch (_context25.prev = _context25.next) {
                case 0:
                  _context25.next = 2;
                  return crypto.verifyDocumentSignature(signed.document, signed.signature, publicSignKey);

                case 2:
                  verified = _context25.sent;
                  return _context25.abrupt('return', _promise2.default.resolve(verified));

                case 4:
                case 'end':
                  return _context25.stop();
              }
            }
          }, _callee25, this);
        }));

        function verify(_x66, _x67) {
          return _ref25.apply(this, arguments);
        }

        return verify;
      }()

      /**
       * Update a record, with optimistic concurrent locking, that already exists in the E3DB system.
       *
       * @param {Record} record Record to be updated.
       *
       * @returns {Promise<Record>} Updated record
       */

    }, {
      key: 'update',
      value: function () {
        var _ref26 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee26(record) {
          var _this = this;

          var recordId, version, recordInfo, encrypted;
          return _regenerator2.default.wrap(function _callee26$(_context26) {
            while (1) {
              switch (_context26.prev = _context26.next) {
                case 0:
                  recordId = record.meta.recordId;
                  version = record.meta.version;

                  // Update record signature

                  recordInfo = new _recordInfo2.default(record.meta, record.data);

                  if (!(this.config.version > 1)) {
                    _context26.next = 9;
                    break;
                  }

                  _context26.next = 6;
                  return this.sign(recordInfo);

                case 6:
                  _context26.t0 = _context26.sent;
                  _context26.next = 10;
                  break;

                case 9:
                  _context26.t0 = null;

                case 10:
                  record.signature = _context26.t0;
                  _context26.next = 13;
                  return encryptRecord(this, record, crypto);

                case 13:
                  encrypted = _context26.sent;
                  return _context26.abrupt('return', oauthFetch(this, this.config.apiUrl + '/v1/storage/records/safe/' + recordId + '/' + version, {
                    method: 'PUT',
                    headers: {
                      'Content-Type': 'application/json'
                    },
                    body: encrypted.stringify()
                  }).then(checkStatus).then(function (response) {
                    return response.json();
                  }).then(_record2.default.decode).then(function (rec) {
                    return rec;
                  }).then(function (record) {
                    return decryptRecord(_this, record, crypto);
                  }));

                case 15:
                case 'end':
                  return _context26.stop();
              }
            }
          }, _callee26, this);
        }));

        function update(_x68) {
          return _ref26.apply(this, arguments);
        }

        return update;
      }()

      /**
       * Deletes a record from the E3DB system
       *
       * @param {string} recordId  ID of the record to remove
       * @param {string} [version] Optional version ID to remove safely
       *
       * @returns {Promise<bool>}
       */

    }, {
      key: 'delete',
      value: function () {
        var _ref27 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee27(recordId) {
          var version = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
          var url, response;
          return _regenerator2.default.wrap(function _callee27$(_context27) {
            while (1) {
              switch (_context27.prev = _context27.next) {
                case 0:
                  url = this.config.apiUrl + '/v1/storage/records/' + recordId;

                  if (version !== null) {
                    url = this.config.apiUrl + '/v1/storage/records/safe/' + recordId + '/' + version;
                  }

                  _context27.next = 4;
                  return oauthFetch(this, url, {
                    method: 'DELETE',
                    headers: {
                      'Content-Type': 'application/json'
                    }
                  });

                case 4:
                  response = _context27.sent;
                  _context27.t0 = response.status;
                  _context27.next = _context27.t0 === 204 ? 8 : _context27.t0 === 403 ? 8 : _context27.t0 === 409 ? 9 : 10;
                  break;

                case 8:
                  return _context27.abrupt('return', _promise2.default.resolve(true));

                case 9:
                  throw new Error('Conflict');

                case 10:
                  throw new Error('Error while deleting record data!');

                case 11:
                case 'end':
                  return _context27.stop();
              }
            }
          }, _callee27, this);
        }));

        function _delete(_x70) {
          return _ref27.apply(this, arguments);
        }

        return _delete;
      }()

      /**
       * Back up the client's configuration to E3DB in a serialized format that can be read
       * by the Admin Console. The stored configuration will be shared with the specified client,
       * and the account service notified that the sharing has taken place.
       *
       * @param {string} clientId          Unique ID of the client to which we're backing up
       * @param {string} registrationToken Original registration token used to create the client
       *
       * @returns {Promise<bool>}
       */

    }, {
      key: 'backup',
      value: function () {
        var _ref28 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee28(clientId, registrationToken) {
          var credentials;
          return _regenerator2.default.wrap(function _callee28$(_context28) {
            while (1) {
              switch (_context28.prev = _context28.next) {
                case 0:
                  /* eslint-disable camelcase */
                  credentials = {
                    version: '"' + this.config.version.toString() + '"',
                    client_id: '"' + this.config.clientId + '"',
                    api_key_id: '"' + this.config.apiKeyId + '"',
                    api_secret: '"' + this.config.apiSecret + '"',
                    client_email: '""',
                    public_key: '"' + this.config.publicKey + '"',
                    private_key: '"' + this.config.privateKey + '"'
                  };

                  if (this.config.version === 2) {
                    credentials.public_signing_key = '"' + this.config.publicSigningKey + '"';
                    credentials.private_signing_key = '"' + this.config.privateSigningKey + '"';
                  }

                  credentials.api_url = '"' + this.config.apiUrl + '"';
                  /* eslint-enable */
                  _context28.next = 5;
                  return this.write('tozny.key_backup', credentials, {
                    client: this.config.clientId
                  });

                case 5:
                  _context28.next = 7;
                  return this.share('tozny.key_backup', clientId);

                case 7:
                  _context28.next = 9;
                  return fetch(this.config.apiUrl + '/v1/account/backup/' + registrationToken + '/' + this.config.clientId, {
                    method: 'POST'
                  });

                case 9:
                  return _context28.abrupt('return', _promise2.default.resolve(true));

                case 10:
                case 'end':
                  return _context28.stop();
              }
            }
          }, _callee28, this);
        }));

        function backup(_x71, _x72) {
          return _ref28.apply(this, arguments);
        }

        return backup;
      }()

      /**
       * Query E3DB records according to a set of selection criteria.
       *
       * The default behavior is to return all records written by the
       * current authenticated client.
       *
       * To restrict the results to a particular type, pass a type or
       * list of types as the `type` argument.
       *
       * To restrict the results to a set of clients, pass a single or
       * list of client IDs as the `writer` argument. To list records
       * written by any client that has shared with the current client,
       * pass the special string 'all' as the `writer` argument.
       *
       * @param {bool}         data     Flag to include data in records
       * @param {string|array} writer   Select records written by a single writer, a list of writers, or 'all'
       * @param {string|array} record   Select a single record or list of records
       * @param {string|array} type     Select records of a single type or a list of types
       * @param {array}        plain    Associative array of plaintext meta to use as a filter
       * @param {number}       pageSize Number of records to fetch per request
       *
       * @returns {QueryResult}
       */

    }, {
      key: 'query',
      value: function query() {
        var data = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : true;
        var writer = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
        var record = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
        var type = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
        var plain = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : null;
        var pageSize = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : DEFAULT_QUERY_COUNT;

        var allWriters = false;
        if (writer === 'all') {
          allWriters = true;
          writer = [];
        }

        var query = new _query3.default(0, data, writer, record, type, plain, null, pageSize, allWriters);
        return new _queryResult2.default(this, query, crypto);
      }

      /**
       * Internal-only method to execute a query against the server and parse the response.
       *
       * @param {Query} query Query request to execute against the server
       *
       * @returns {QueryResult}
       */

    }, {
      key: '_query',
      value: function () {
        var _ref29 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee29(query) {
          var response;
          return _regenerator2.default.wrap(function _callee29$(_context29) {
            while (1) {
              switch (_context29.prev = _context29.next) {
                case 0:
                  _context29.next = 2;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/search', {
                    method: 'POST',
                    headers: {
                      'Content-Type': 'application/json'
                    },
                    body: query.stringify()
                  });

                case 2:
                  response = _context29.sent;
                  _context29.next = 5;
                  return checkStatus(response);

                case 5:
                  return _context29.abrupt('return', response.json());

                case 6:
                case 'end':
                  return _context29.stop();
              }
            }
          }, _callee29, this);
        }));

        function _query(_x79) {
          return _ref29.apply(this, arguments);
        }

        return _query;
      }()

      /**
       * Grant another E3DB client access to records of a particular type.
       *
       * @param {string} type     Type of records to share
       * @param {string} readerId Client ID or email address of reader to grant access to
       *
       * @returns {Promise<bool>}
       */

    }, {
      key: 'share',
      value: function () {
        var _ref30 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee30(type, readerId) {
          var clientInfo, clientId, ak, policy, request;
          return _regenerator2.default.wrap(function _callee30$(_context30) {
            while (1) {
              switch (_context30.prev = _context30.next) {
                case 0:
                  if (!(readerId === this.config.clientId)) {
                    _context30.next = 2;
                    break;
                  }

                  return _context30.abrupt('return', _promise2.default.resolve(true));

                case 2:
                  if (!EMAIL.test(readerId)) {
                    _context30.next = 7;
                    break;
                  }

                  _context30.next = 5;
                  return this.clientInfo(readerId);

                case 5:
                  clientInfo = _context30.sent;
                  return _context30.abrupt('return', this.share(type, clientInfo.clientId));

                case 7:
                  clientId = this.config.clientId;
                  _context30.next = 10;
                  return getAccessKey(this, clientId, clientId, clientId, type, crypto);

                case 10:
                  ak = _context30.sent;

                  if (!(ak === null)) {
                    _context30.next = 17;
                    break;
                  }

                  _context30.next = 14;
                  return crypto.randomKey();

                case 14:
                  ak = _context30.sent;
                  _context30.next = 17;
                  return putAccessKey(this, this.config.clientId, this.config.clientId, this.config.clientId, type, ak);

                case 17:
                  _context30.next = 19;
                  return putAccessKey(this, clientId, clientId, readerId, type, ak, crypto);

                case 19:
                  policy = { allow: [{ read: {} }] };
                  _context30.next = 22;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/policy/' + clientId + '/' + clientId + '/' + readerId + '/' + type, {
                    method: 'PUT',
                    headers: {
                      'Content-Type': 'application/json'
                    },
                    body: (0, _stringify2.default)(policy)
                  });

                case 22:
                  request = _context30.sent;
                  _context30.next = 25;
                  return checkStatus(request);

                case 25:
                  return _context30.abrupt('return', _promise2.default.resolve(true));

                case 26:
                case 'end':
                  return _context30.stop();
              }
            }
          }, _callee30, this);
        }));

        function share(_x80, _x81) {
          return _ref30.apply(this, arguments);
        }

        return share;
      }()

      /**
       * Revoke another E3DB client's access to records of a particular type.
       *
       * @param {string} type     Type of records to share
       * @param {string} readerId Client ID or email address of reader to grant access from
       *
       * @returns {Promise<bool>}
       */

    }, {
      key: 'revoke',
      value: function () {
        var _ref31 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee31(type, readerId) {
          var clientInfo, clientId, policy, request;
          return _regenerator2.default.wrap(function _callee31$(_context31) {
            while (1) {
              switch (_context31.prev = _context31.next) {
                case 0:
                  if (!(readerId === this.config.clientId)) {
                    _context31.next = 2;
                    break;
                  }

                  return _context31.abrupt('return', _promise2.default.resolve(true));

                case 2:
                  if (!EMAIL.test(readerId)) {
                    _context31.next = 7;
                    break;
                  }

                  _context31.next = 5;
                  return this.clientInfo(readerId);

                case 5:
                  clientInfo = _context31.sent;
                  return _context31.abrupt('return', this.revoke(type, clientInfo.clientId));

                case 7:
                  clientId = this.config.clientId;
                  policy = { deny: [{ read: {} }] };
                  _context31.next = 11;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/policy/' + clientId + '/' + clientId + '/' + readerId + '/' + type, {
                    method: 'PUT',
                    headers: {
                      'Content-Type': 'application/json'
                    },
                    body: (0, _stringify2.default)(policy)
                  });

                case 11:
                  request = _context31.sent;
                  _context31.next = 14;
                  return checkStatus(request);

                case 14:
                  _context31.next = 16;
                  return deleteAccessKey(this, clientId, clientId, readerId, type);

                case 16:
                  return _context31.abrupt('return', _promise2.default.resolve(true));

                case 17:
                case 'end':
                  return _context31.stop();
              }
            }
          }, _callee31, this);
        }));

        function revoke(_x82, _x83) {
          return _ref31.apply(this, arguments);
        }

        return revoke;
      }()

      /**
       * Get a list of all outgoing sharing policy relationships
       *
       * @returns {Promise<array>}
       */

    }, {
      key: 'outgoingSharing',
      value: function () {
        var _ref32 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee32() {
          var request, response, json;
          return _regenerator2.default.wrap(function _callee32$(_context32) {
            while (1) {
              switch (_context32.prev = _context32.next) {
                case 0:
                  _context32.next = 2;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/policy/outgoing', {
                    method: 'GET',
                    headers: {
                      'Content-Type': 'application/json'
                    }
                  });

                case 2:
                  request = _context32.sent;
                  _context32.next = 5;
                  return checkStatus(request);

                case 5:
                  response = _context32.sent;
                  _context32.next = 8;
                  return response.json();

                case 8:
                  json = _context32.sent;
                  return _context32.abrupt('return', _promise2.default.all(json.map(_outgoingSharingPolicy2.default.decode)));

                case 10:
                case 'end':
                  return _context32.stop();
              }
            }
          }, _callee32, this);
        }));

        function outgoingSharing() {
          return _ref32.apply(this, arguments);
        }

        return outgoingSharing;
      }()

      /**
       * Get a list of all incoming sharing policy relationships
       *
       * @returns {Promise<array>}
       */

    }, {
      key: 'incomingSharing',
      value: function () {
        var _ref33 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee33() {
          var request, response, json;
          return _regenerator2.default.wrap(function _callee33$(_context33) {
            while (1) {
              switch (_context33.prev = _context33.next) {
                case 0:
                  _context33.next = 2;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/policy/incoming', {
                    method: 'GET',
                    headers: {
                      'Content-Type': 'application/json'
                    }
                  });

                case 2:
                  request = _context33.sent;
                  _context33.next = 5;
                  return checkStatus(request);

                case 5:
                  response = _context33.sent;
                  _context33.next = 8;
                  return response.json();

                case 8:
                  json = _context33.sent;
                  return _context33.abrupt('return', _promise2.default.all(json.map(_incomingSharingPolicy2.default.decode)));

                case 10:
                case 'end':
                  return _context33.stop();
              }
            }
          }, _callee33, this);
        }));

        function incomingSharing() {
          return _ref33.apply(this, arguments);
        }

        return incomingSharing;
      }()
    }, {
      key: 'writeLargeFile',
      value: function () {
        var _ref34 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee34(recordType, fileObject) {
          var plainMetadata = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

          var clientId, ak, _ref35, _ref36, encryptedFile, checkSum, encryptedLength, fileCompression, fileObj, postBody, postRequest, postResponse, postJson, putRequest, patchRequest, patchResponse, patchJson;

          return _regenerator2.default.wrap(function _callee34$(_context34) {
            while (1) {
              switch (_context34.prev = _context34.next) {
                case 0:
                  clientId = this.config.clientId;
                  _context34.next = 3;
                  return getAccessKey(this, clientId, clientId, clientId, recordType, crypto);

                case 3:
                  ak = _context34.sent;

                  if (!(ak === null)) {
                    _context34.next = 10;
                    break;
                  }

                  _context34.next = 7;
                  return crypto.randomKey();

                case 7:
                  ak = _context34.sent;
                  _context34.next = 10;
                  return putAccessKey(this, clientId, clientId, clientId, recordType, ak, crypto);

                case 10:
                  _context34.next = 12;
                  return crypto.encryptLargeFile(fileObject, ak, plainMetadata);

                case 12:
                  _ref35 = _context34.sent;
                  _ref36 = (0, _slicedToArray3.default)(_ref35, 3);
                  encryptedFile = _ref36[0];
                  checkSum = _ref36[1];
                  encryptedLength = _ref36[2];
                  fileCompression = 'raw';
                  fileObj = new _file2.default(checkSum, fileCompression, encryptedLength, this.config.clientId, this.config.clientId, recordType, plainMetadata);
                  postBody = fileObj.toJson();
                  // Post file meta to request AWS S3 Bucket URL.

                  _context34.next = 22;
                  return oauthFetch(this, this.config.apiUrl + '/v1/storage/files', {
                    method: 'POST',
                    headers: {
                      'Content-Type': 'application/json'
                    },
                    body: (0, _stringify2.default)(postBody)
                  });

                case 22:
                  postRequest = _context34.sent;
                  _context34.next = 25;
                  return checkStatus(postRequest);

                case 25:
                  postResponse = _context34.sent;
                  _context34.next = 28;
                  return postResponse.json();

                case 28:
                  postJson = _context34.sent;

                  fileObj.fileUrl = postJson.file_url;
                  fileObj.recordId = postJson.id;
                  // Upload encrypted file to AWS S3 Bucket.
                  _context34.next = 33;
                  return fetch(fileObj.fileUrl, {
                    method: 'PUT',
                    headers: {
                      'Content-Type': 'application/octet-stream',
                      'Content-MD5': checkSum
                    },
                    body: encryptedFile
                  });

                case 33:
                  putRequest = _context34.sent;
                  _context34.next = 36;
                  return checkStatus(putRequest);

                case 36:
                  _context34.next = 38;
                  return oauthFetch(this, this.config.apiUrl + ('/v1/storage/files/' + fileObj.recordId), {
                    method: 'PATCH',
                    headers: {
                      'Content-Type': 'application/json'
                    }
                  });

                case 38:
                  patchRequest = _context34.sent;
                  _context34.next = 41;
                  return checkStatus(patchRequest);

                case 41:
                  patchResponse = _context34.sent;
                  _context34.next = 44;
                  return patchResponse.json();

                case 44:
                  patchJson = _context34.sent;
                  return _context34.abrupt('return', new _file2.default(patchJson.meta.file_meta.checksum, patchJson.meta.file_meta.compression, patchJson.meta.file_meta.size, patchJson.meta.writer_id, patchJson.meta.user_id, patchJson.meta.type, patchJson.meta.plain, patchJson.meta.file_meta.file_url, patchJson.meta.file_meta.file_name, patchJson.meta.record_id, patchJson.meta.created, patchJson.meta.last_modified, patchJson.meta.version));

                case 46:
                case 'end':
                  return _context34.stop();
              }
            }
          }, _callee34, this);
        }));

        function writeLargeFile(_x85, _x86) {
          return _ref34.apply(this, arguments);
        }

        return writeLargeFile;
      }()

      // The file_url appears to be null.  Is that right?

    }, {
      key: 'readLargeFile',
      value: function () {
        var _ref37 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee35(recordId, destinationFilename) {
          var getRequest, getResponse, getJson, fileObj, ak, fileRequest, fileResponse, encryptedFileArrayBuffer, fileName, decrypted;
          return _regenerator2.default.wrap(function _callee35$(_context35) {
            while (1) {
              switch (_context35.prev = _context35.next) {
                case 0:
                  _context35.next = 2;
                  return oauthFetch(this, this.config.apiUrl + ('/v1/storage/files/' + recordId), {
                    method: 'GET',
                    headers: {
                      'Content-Type': 'application/json'
                    }
                  });

                case 2:
                  getRequest = _context35.sent;
                  _context35.next = 5;
                  return checkStatus(getRequest);

                case 5:
                  getResponse = _context35.sent;
                  _context35.next = 8;
                  return getResponse.json();

                case 8:
                  getJson = _context35.sent;

                  // Create File instance.
                  fileObj = new _file2.default(getJson.meta.file_meta.checksum, getJson.meta.file_meta.compression, getJson.meta.file_meta.size, getJson.meta.writer_id, getJson.meta.user_id, getJson.meta.type, getJson.meta.plain, getJson.meta.file_meta.file_url, getJson.meta.file_meta.file_name, getJson.meta.record_id, getJson.meta.created, getJson.meta.last_modified, getJson.meta.version);
                  // Get the access key to decrypt the record.

                  _context35.next = 12;
                  return getAccessKey(this, fileObj.writerId, this.config.clientId, fileObj.writerId, fileObj.recordType, crypto);

                case 12:
                  ak = _context35.sent;

                  if (!(ak === null)) {
                    _context35.next = 15;
                    break;
                  }

                  throw new Error('Can\'t read records of type ' + fileObj.type);

                case 15:
                  _context35.next = 17;
                  return fetch(fileObj.fileUrl, {
                    method: 'GET'
                  });

                case 17:
                  fileRequest = _context35.sent;
                  _context35.next = 20;
                  return checkStatus(fileRequest);

                case 20:
                  fileResponse = _context35.sent;
                  _context35.next = 23;
                  return fileResponse.arrayBuffer();

                case 23:
                  encryptedFileArrayBuffer = _context35.sent;

                  // Should take the encryptedFileName and destinationFileName
                  fileName = destinationFilename ? destinationFilename : getJson.meta.file_meta.file_name;
                  _context35.next = 27;
                  return crypto.decryptFile(fileName, ak, encryptedFileArrayBuffer);

                case 27:
                  decrypted = _context35.sent;
                  return _context35.abrupt('return', [decrypted, fileObj]);

                case 29:
                case 'end':
                  return _context35.stop();
              }
            }
          }, _callee35, this);
        }));

        function readLargeFile(_x87, _x88) {
          return _ref37.apply(this, arguments);
        }

        return readLargeFile;
      }()
    }]);
    return Client;
  }();

  return Client;
};
exports.default = PreClient;