/*!
 * Tozny e3db
 *
 * LICENSE
 *
 * Tozny dual licenses this product. For commercial use, please contact
 * info@tozny.com. For non-commercial use, this license permits use of the
 * software only by government agencies, schools, universities, non-profit
 * organizations or individuals on projects that do not receive external
 * funding other than government research grants and contracts. Any other use
 * requires a commercial license. For the full license, please see LICENSE.md,
 * in this source repository.
 *
 * @copyright Copyright (c) 2018-19 Tozny, LLC (https://tozny.com)
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _getPrototypeOf = require('babel-runtime/core-js/object/get-prototype-of');

var _getPrototypeOf2 = _interopRequireDefault(_getPrototypeOf);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _possibleConstructorReturn2 = require('babel-runtime/helpers/possibleConstructorReturn');

var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);

var _inherits2 = require('babel-runtime/helpers/inherits');

var _inherits3 = _interopRequireDefault(_inherits2);

var _publicKey = require('./publicKey');

var _publicKey2 = _interopRequireDefault(_publicKey);

var _serializable = require('../types/serializable');

var _serializable2 = _interopRequireDefault(_serializable);

var _signingKey = require('./signingKey');

var _signingKey2 = _interopRequireDefault(_signingKey);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Representation of a cached, encrypted access key
 *
 * @property {string} eak
 * @property {string} authorizerID
 * @property {string} authorizerPublicKey
 * @property {string} signerId
 * @property {string} signerSigningKey
 */
var EAKInfo = function (_Serializable) {
  (0, _inherits3.default)(EAKInfo, _Serializable);

  function EAKInfo(eak, authorizerID, authorizerPublicKey, signerId, signerSigningKey) {
    (0, _classCallCheck3.default)(this, EAKInfo);

    var _this = (0, _possibleConstructorReturn3.default)(this, (EAKInfo.__proto__ || (0, _getPrototypeOf2.default)(EAKInfo)).call(this));

    _this.eak = eak;
    _this.authorizerID = authorizerID;
    _this.authorizerPublicKey = new _publicKey2.default(authorizerPublicKey);
    _this.signerId = signerId;
    _this.signerSigningKey = new _signingKey2.default(signerSigningKey);
    return _this;
  }

  /* eslint-disable camelcase */

  /**
   * Generate a JSON.stringify-friendly version of the object
   * automatically omitting any `null` fields.
   *
   * @returns {object}
   */


  (0, _createClass3.default)(EAKInfo, [{
    key: 'serializable',
    value: function serializable() {
      var toSerialize = {
        eak: this.eak,
        authorizer_id: this.authorizerID,
        authorizer_public_key: this.authorizerPublicKey,
        signer_id: this.signerId,
        signer_signing_key: this.signerSigningKey
      };

      for (var key in toSerialize) {
        if (toSerialize.hasOwnProperty(key)) {
          if (toSerialize[key] === null) {
            delete toSerialize[key];
          }
        }
      }

      return toSerialize;
    }

    /* eslint-enable */

    /**
     * Specify how an already unserialized JSON array should be marshaled into
     * an object representation.
     *
     * The array expected for deserializing back into an object requires:
     *
     * <code>
     * $eakInfo = EAKInfo::decode({
     *   'eak'                   => '',
     *   'authorizer_id'         => '',
     *   'authorizer_public_key' => '',
     *   'signer_id'             => '',
     *   'signer_signing_key'    => ''
     * });
     * </code>
     *
     * @param {object} json
     *
     * @return {Promise<EAKInfo>}
     */

  }], [{
    key: 'decode',
    value: function () {
      var _ref = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee(json) {
        var signerId, signingKey;
        return _regenerator2.default.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                signerId = null;
                signingKey = null;

                if (json.hasOwnProperty('signer_id')) {
                  signerId = json.signer_id;
                }
                if (json.hasOwnProperty('signer_signing_key') && json.signer_signing_key !== null && json.signer_signing_key.hasOwnProperty('ed25519')) {
                  signingKey = json.signer_signing_key.ed25519;
                }

                return _context.abrupt('return', _promise2.default.resolve(new EAKInfo(json.eak, json.authorizer_id, json.authorizer_public_key.curve25519, signerId, signingKey)));

              case 5:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function decode(_x) {
        return _ref.apply(this, arguments);
      }

      return decode;
    }()
  }]);
  return EAKInfo;
}(_serializable2.default);

exports.default = EAKInfo;