/*!
 * Tozny e3db
 *
 * LICENSE
 *
 * Tozny dual licenses this product. For commercial use, please contact
 * info@tozny.com. For non-commercial use, the contents of this file are
 * subject to the TOZNY NON-COMMERCIAL LICENSE (the "License") which
 * permits use of the software only by government agencies, schools,
 * universities, non-profit organizations or individuals on projects that
 * do not receive external funding other than government research grants
 * and contracts.  Any other use requires a commercial license. You may
 * not use this file except in compliance with the License. You may obtain
 * a copy of the License at https://tozny.com/legal/non-commercial-license.
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License. Portions of the software are Copyright (c) TOZNY LLC, 2018-19.
 * All rights reserved.
 *
 * @copyright Copyright (c) 2018-19 Tozny, LLC (https://tozny.com)
 */

'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _getPrototypeOf = require('babel-runtime/core-js/object/get-prototype-of');

var _getPrototypeOf2 = _interopRequireDefault(_getPrototypeOf);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _possibleConstructorReturn2 = require('babel-runtime/helpers/possibleConstructorReturn');

var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);

var _inherits2 = require('babel-runtime/helpers/inherits');

var _inherits3 = _interopRequireDefault(_inherits2);

var _meta = require('./meta');

var _meta2 = _interopRequireDefault(_meta);

var _recordData = require('./recordData');

var _recordData2 = _interopRequireDefault(_recordData);

var _signable = require('./signable');

var _signable2 = _interopRequireDefault(_signable);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * A E3DB record containing data and metadata. Records are
 * a key/value mapping containing data serialized
 * into strings. All records are encrypted prior to sending them
 * to the server for storage, and decrypted in the client after
 * they are read.
 *
 * @property {Meta}       meta      Meta information about the record.
 * @property {RecordData} data      Either plaintext or encrypted record fields
 * @property {string}     signature Signature over unencrypted record data
 */
var Record = function (_Signable) {
  (0, _inherits3.default)(Record, _Signable);

  function Record(meta, data) {
    var signature = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    (0, _classCallCheck3.default)(this, Record);

    var _this = (0, _possibleConstructorReturn3.default)(this, (Record.__proto__ || (0, _getPrototypeOf2.default)(Record)).call(this));

    if (meta instanceof _meta2.default) {
      _this.meta = meta;
    } else {
      throw new Error('Record meta must be a Meta object!');
    }

    if (data instanceof _recordData2.default || data === null) {
      _this.data = data;
    } else if (data instanceof Object) {
      _this.data = new _recordData2.default(data);
    } else {
      throw new Error('Record data must be an object!');
    }

    _this.signature = signature;
    return _this;
  }

  /* eslint-disable camelcase */

  /**
   * Generate a JSON.stringify-friendly version of the object
   * automatically omitting any `null` fields.
   *
   * @returns {object}
   */


  (0, _createClass3.default)(Record, [{
    key: 'serializable',
    value: function serializable() {
      return {
        meta: this.meta.serializable(),
        data: this.data.serializable(),
        rec_sig: this.signature
      };
    }

    /* eslint-enabled */

    /**
     * Specify how an already unserialized JSON array should be marshaled into
     * an object representation.
     *
     * Records consist of two elements, meta and data. The array we deserialize into a Record instance
     * must match this format. The meta element is itself an array representing the Meta class. The
     * data element is a simpler array mapping string keys to either encrypted or plaintext string values.
     *
     * <code>
     * record = Record::decode({
     *   meta: {
     *     record_id:     '',
     *     writer_id:     '',
     *     user_id:       '',
     *     type:          '',
     *     plain:         {},
     *     created:       '',
     *     last_modified: '',
     *     version:       ''
     *   },
     *   data: {
     *     key1: 'value',
     *     key2: 'value'
     *   },
     *   rec_sig: ''
     * })
     * </code>
     *
     * @param {array} parsed
     *
     * @return {Promise<Record>}
     */

  }], [{
    key: 'decode',
    value: function () {
      var _ref = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee(json) {
        var meta, signature;
        return _regenerator2.default.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _meta2.default.decode(json.meta);

              case 2:
                meta = _context.sent;
                signature = json.rec_sig === undefined ? null : json.rec_sig;
                return _context.abrupt('return', _promise2.default.resolve(new Record(meta, json.data, signature)));

              case 5:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function decode(_x2) {
        return _ref.apply(this, arguments);
      }

      return decode;
    }()
  }]);
  return Record;
}(_signable2.default);

exports.default = Record;